package com.abmyers.spinfw.servlet;

import java.io.IOException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import com.abmyers.spinfw.action.ActionContext;
import com.abmyers.utils.Framework;
import com.abmyers.utils.StringUtils;

/**
 *
 * @author amyers
 */
public class SpinFWServlet extends HttpServlet {

	/**
	 *
	 */
	private static final long serialVersionUID = 1L;
	private String packagePrefix = "";
	private String actionPath = "/actions";
	private HashMap<String, Object> controllers = null;
	public final static Logger logger = Logger.getLogger(SpinFWServlet.class
			.getName());

	@Override
	public void init() throws ServletException {
		controllers = new HashMap<String, Object>();
		this.packagePrefix = getServletConfig().getInitParameter(
				"packagePrefix");
		if (this.packagePrefix != null && !this.packagePrefix.endsWith(".")) {
			this.packagePrefix = this.packagePrefix + ".";
		} else if (this.packagePrefix == null) {
			this.packagePrefix = "";
		}

		if (getServletConfig().getInitParameter("actionPath") != null) {
			this.actionPath = getServletConfig().getInitParameter("actionPath");
		}

		this.getServletContext().setAttribute("actionPath", this.actionPath);

		logger.log(Level.FINE, "packagePrefix set to {0}", this.packagePrefix);
	}

	/**
	 * Processes requests for both HTTP <code>GET</code> and <code>POST</code>
	 * methods.
	 *
	 * @param request
	 *            servlet request
	 * @param response
	 *            servlet response
	 * @throws ServletException
	 *             if a servlet-specific error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	@SuppressWarnings("rawtypes")
	protected void processRequest(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {

		String[] splitString = request.getPathInfo().split("/");
		String actionString = (splitString == null) ? null : splitString[1];
		if (actionString == null) {
			throw new IllegalArgumentException("Parameter named action must "
					+ "be provided");
		}

		String shortControllerClassName = StringUtils.convertSnakeToCamelCase(
				actionString, true);
		String fullControllerClassName = packagePrefix + "controllers."
				+ StringUtils.convertSnakeToCamelCase(actionString, true)
				+ "Controller";
		String controllerMethodName = "index"; // default value

		if (splitString.length > 2) {
			controllerMethodName = splitString[2];
		}

		logger.log(Level.FINE, "packagePrefix {0}", packagePrefix);
		logger.log(Level.FINE, "controllerClassName {0}",
				shortControllerClassName);
		logger.log(Level.FINE, "controllerMethodName {0}", controllerMethodName);

		try {

			ActionContext ac = new ActionContext(
					shortControllerClassName.toLowerCase() + "."
							+ controllerMethodName.toLowerCase(), request);
			if (isControllerExists(fullControllerClassName)) {
				// call any controller method
				Class c = Class.forName(fullControllerClassName);

				Object obj = controllers.get(fullControllerClassName);

				try {
					@SuppressWarnings("unchecked")
					Method method = c.getMethod(controllerMethodName,
							ActionContext.class);

					method.invoke(obj, ac);
				} catch (NoSuchMethodException nsme) {
					// method doesn't exist, so nothing to do
				}
			}

			request.setAttribute("rc", ac.getRc());

			if (ac.getRedirectTo() != null) {
				response.sendRedirect(Framework.createHref(request.getContextPath(), this.actionPath, ac.getRedirectTo()));
				return;
			}

			/*
			 * Look for 3 levels of layout. - layouts/section/item.jsp: The
			 * view-specific layout (required) - layouts/section.jsp: The
			 * section-specific layout (optional) - layouts/default.jsp: The
			 * site-wide layout (required)
			 */

			String viewSpecificLayout = "/WEB-INF/views/" + ac.getViewSection()
					+ "/" + ac.getViewItem() + ".jsp";
			String sectionSpecificLayout = "/WEB-INF/layouts/"
					+ ac.getViewSection() + ".jsp";
			String defaultLayout = "/WEB-INF/layouts/default.jsp";

			if (getServletContext().getResource(sectionSpecificLayout) != null) {
				request.setAttribute("viewSpecificLayout", viewSpecificLayout);
				request.setAttribute("sectionSpecificLayout",
						sectionSpecificLayout);
			} else {
				request.setAttribute("viewSpecificLayout", viewSpecificLayout);
				request.setAttribute("sectionSpecificLayout", null);
			}

			request.getRequestDispatcher(defaultLayout).include(request,
					response);

		} catch (Exception ex) {
			logger.log(Level.SEVERE, null, ex);
			throw new ServletException(ex.getMessage(), ex);
		}
	}

	/**
	 * Handles the HTTP <code>GET</code> method.
	 *
	 * @param request
	 *            servlet request
	 * @param response
	 *            servlet response
	 * @throws ServletException
	 *             if a servlet-specific error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	@Override
	protected void doGet(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		processRequest(request, response);
	}

	/**
	 * Handles the HTTP <code>POST</code> method.
	 *
	 * @param request
	 *            servlet request
	 * @param response
	 *            servlet response
	 * @throws ServletException
	 *             if a servlet-specific error occurs
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	@Override
	protected void doPost(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		processRequest(request, response);
	}

	/**
	 * Returns a short description of the servlet.
	 *
	 * @return a String containing servlet description
	 */
	@Override
	public String getServletInfo() {
		return "Short description";
	}

	private boolean isControllerExists(String controllerClassName) {
    	if (controllers.containsKey(controllerClassName)) {
    		return true;
    	} else {
	        boolean result;
	        try {
	        	@SuppressWarnings("rawtypes")
	            Class c = Class.forName(controllerClassName);
	            // it exists on the classpath
	            controllers.put(controllerClassName, c.newInstance());
	            result = true;
	        } catch (ClassNotFoundException e) {
	        	logger.log(Level.INFO, "controller " + controllerClassName + " not on classpath");
	            result = false;
	        } catch (InstantiationException e) {
	        	result = false;
			} catch (IllegalAccessException e) {
				result = false;
			}
	        return result;
    	}
	}
}
